// SPDX-License-Identifier: GPL-2.0
/*
 * File:	board.c
 * Description:	BFK3 and BFK1.6+ Baikal-T1 based board initialization
 * Authors:
 *
 * Copyright 2019-2020, Baikal Electronics JSC
 */

#ifdef CONFIG_BOARD_LATE_INIT

#include <common.h>
#include <scsi.h>
#include <usb.h>
#include <stdio_dev.h>
#include <asm/gpio.h>
#include <spi.h>
#include <spi_flash.h>
#include <env.h>
#include <dm.h>
#include <i2c_eeprom.h>
#include <i2c.h>
#include <net.h>

DECLARE_GLOBAL_DATA_PTR;

#define BOOT_FLASH_ID_LEN	20
#define CONFIG_BAIKAL_MAC_BASE	0x4CA515000000ULL

#define ETH_ID_1G_1	1
#define ETH_ID_1G_2	2
#define ETH_ID_10G	0

#ifdef CONFIG_BOARD_BFKX
struct vendor_data {
	u32 crc;		/* Reserved for CRC */
	u8 type;		/* Board ID */
	u8 revision;	/* Board revision in BCD format */
	u64 serial;		/* Board Serial number */
	u16 cpu;		/* CPU serial number */
};

uint8_t eeprom_buf[BAIKAL_EEPROM_SIZE];
struct vendor_data *vdata;
#endif

void calc_and_set_macaddr(int serial)
{
	u64 macaddr;
	u8* b = (u8*)&macaddr;
	uint8_t enetaddr[6];

	macaddr = CONFIG_BAIKAL_MAC_BASE | ((serial << 2) & 0xFFFFFC);
	for (int i = 0; i < 6; i++) enetaddr[i] = b[5 - i];

	/* Ethernet 10G (eth0) */
	enetaddr[5] = b[0] | ETH_ID_10G;
	eth_env_set_enetaddr("ethaddr", enetaddr);

	/* Ethernet 1G (eth1) */
	enetaddr[5] = b[0] | ETH_ID_1G_1;
	eth_env_set_enetaddr("eth1addr", enetaddr);

	/* Ethernet 1G (eth2) */
	enetaddr[5] = b[0] | ETH_ID_1G_2;
	eth_env_set_enetaddr("eth2addr", enetaddr);
}

unsigned long get_spi_chip_id(void)
{
	int spi_get_bus_error, spi_claim_error, spi_read_error;
	struct spi_slave *slave;
	struct udevice *bus;
	uint8_t data[BOOT_FLASH_ID_LEN + 1];
	const char cmd[BOOT_FLASH_ID_LEN + 1] = {0x9F};
	char flash_chip_id[BOOT_FLASH_ID_LEN * 2 + 1];

	flash_chip_id[BOOT_FLASH_ID_LEN * 2] = '\0';

	spi_get_bus_error = spi_get_bus_and_cs(CONFIG_SF_DEFAULT_BUS, CONFIG_SF_DEFAULT_CS, CONFIG_SF_DEFAULT_SPEED, CONFIG_SF_DEFAULT_MODE, NULL, 0, &bus, &slave);
	if (spi_get_bus_error) {
		printf("Error: Failed to get spi bus (%d)\n", spi_get_bus_error);
		return env_get_ulong("cpu_serial", 10, 0);
	}

	spi_claim_error = spi_claim_bus(slave);
	if (spi_claim_error) {
		printf("Error: Failed to claim spi bus (%d)\n", spi_claim_error);
		return env_get_ulong("cpu_serial", 10, 0);
	}

	spi_read_error = spi_xfer(slave, 168, cmd, data, SPI_XFER_BEGIN | SPI_XFER_END);
	spi_release_bus(slave);
	if (spi_read_error) {
		printf("Error: Failed to read boot flash ID (%d)\n", spi_read_error);
		return env_get_ulong("cpu_serial", 10, 0);
	}

	for (int i = 1; i < BOOT_FLASH_ID_LEN + 1; i++)
		sprintf(flash_chip_id + (i - 1) * 2, "%02x", data[i]);
	printf("RomID: %s\n", flash_chip_id);

	return simple_strtol(flash_chip_id + (BOOT_FLASH_ID_LEN - 6), NULL, 16);
}

#ifdef CONFIG_BOARD_BFKX
int set_macaddr_from_eeprom(void)
{
	uint8_t enetaddr[6], old_addr[6];
	char name[32];
	int i, rc, ret = 0;

	for (i = 0; i < BAIKAL_EEPROM_MAX_IFACES; i++) {
		switch (i) {
		case 0:
			memcpy(enetaddr, (uint8_t *)&eeprom_buf[BAIKAL_EEPROM_MAC0_OFFSET], 6);
			break;
		case 1:
			memcpy(enetaddr, (uint8_t *)&eeprom_buf[BAIKAL_EEPROM_MAC1_OFFSET], 6);
			break;
		case 2:
			memcpy(enetaddr, (uint8_t *)&eeprom_buf[BAIKAL_EEPROM_MAC2_OFFSET], 6);
			break;
		default:
			break;
		}

		if (is_valid_ethaddr(enetaddr)) {
			sprintf(name, i ? "eth%daddr" : "ethaddr", i);
			rc = eth_env_get_enetaddr(name, old_addr);
			if (!rc) { /* Address not set - it is safe to set it */
				rc = eth_env_set_enetaddr(name, enetaddr);
				if (rc) {
					printf("%s: Cannot set %s to %pM\n",
						__func__, name, enetaddr);
					ret++;
				}
			} else if (memcmp(enetaddr, old_addr, 6)) {
				/* Address is set but different from EEPROM */
				printf("%s mismatch:\n"
					"Address in EEPROM is       %pM\n"
					"Address in environment is  %pM\n",
					name, enetaddr, old_addr);
				printf("Please fix it manually and save\n");
				ret++;
			}
		} else {
			printf("%s: Address %pM is not valid for iface %d\n",
				__func__, enetaddr, i);
			ret++;
		}
	}

	return ret;
}

int read_eeprom(void)
{
	int eeprom_get_device_error, eeprom_read_error;
	struct udevice *dev;

	eeprom_get_device_error = uclass_get_device_by_name(UCLASS_I2C_EEPROM, BAIKAL_EEPROM_NAME, &dev);
	if (eeprom_get_device_error) {
		printf("Error: Failed to get eeprom device by name (%d)\n", eeprom_get_device_error);
		return -1;
	}

	eeprom_read_error = i2c_eeprom_read(dev, 0, eeprom_buf, BAIKAL_EEPROM_SIZE);
	if (eeprom_read_error) {
		printf("Error: Failed to eeprom read (%d)\n", eeprom_read_error);
		return -1;
	}

	return 0;
}

int read_vendor_data_from_eeprom(void) {
	char buf[32];

	vdata = (struct vendor_data *)&eeprom_buf[BAIKAL_EEPROM_VENDOR_OFFSET];

	printf("Rev:   %u.%u\n", (vdata->revision >> 4) & 0x0f, vdata->revision & 0x0f);
	/* Set board revision */
	snprintf(buf, 10, "%u.%u", (vdata->revision >> 4) & 0x0f, vdata->revision & 0x0f);
	env_set("board_rev", buf);

	if (vdata->serial != 0) {
		printf("S/N:   %llu\n", vdata->serial);
		/* Set board serial number */
		snprintf(buf, 20, "%llu", vdata->serial);
		env_set("board_serial", buf);
	}

	if (vdata->cpu != 0) {
		/* Set CPU serial number */
		snprintf(buf, 10, "%hu", vdata->cpu);
		env_set("cpu_serial", buf);
	}

	return 0;
}
#endif

int board_late_init(void)
{
	unsigned long brd_serial_id;
	char *stdio_name;

#ifdef CONFIG_BOARD_BFK3
	debug("bfk3: board_late_init\n");

	brd_serial_id = get_spi_chip_id();
	calc_and_set_macaddr(brd_serial_id);
#elif defined(CONFIG_BOARD_BFKX)
	debug("bfk1.6+: board_late_init\n");

	if (read_eeprom()) {
		brd_serial_id = get_spi_chip_id();
		calc_and_set_macaddr(brd_serial_id);
	}
	else {
		if (set_macaddr_from_eeprom()) {
			brd_serial_id = get_spi_chip_id();
			calc_and_set_macaddr(brd_serial_id);
		}
		read_vendor_data_from_eeprom();
	}
#endif

	/* Set board name */
	env_set("board_name", BOARD_NAME);

#ifdef CONFIG_DM_PCI
	pci_init();
#ifdef CONFIG_DM_VIDEO
	stdio_name = env_get("stdout");
	if (stdio_name && strstr(stdio_name, "vidconsole"))
		env_set("stdout", stdio_name);
#endif
#endif

#ifdef CONFIG_DM_USB
	usb_init();
#ifdef CONFIG_USB_KEYBOARD
	stdio_name = env_get("stdin");
	if (stdio_name && strstr(stdio_name, "usbkbd"))
		env_set("stdin", stdio_name);
#endif
#endif

#ifdef CONFIG_DM_SCSI
	scsi_scan(false);
#endif

	return 0;
}
#endif
